// ----------------------------------------------------------------------------------
// Microsoft Developer & Platform Evangelism
// 
// Copyright (c) Microsoft Corporation. All rights reserved.
// 
// THIS CODE AND INFORMATION ARE PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND, 
// EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE IMPLIED WARRANTIES 
// OF MERCHANTABILITY AND/OR FITNESS FOR A PARTICULAR PURPOSE.
// ----------------------------------------------------------------------------------
// The example companies, organizations, products, domain names,
// e-mail addresses, logos, people, places, and events depicted
// herein are fictitious.  No association with any real company,
// organization, product, domain name, email address, logo, person,
// places, or events is intended or should be inferred.
// ----------------------------------------------------------------------------------

using System;
using System.Collections.Generic;
using System.Linq;
using System.Net;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Animation;
using System.Windows.Shapes;
using Microsoft.Phone.Controls;
using Microsoft.Phone.Notification;
using System.Diagnostics;
using System.Windows.Threading;
using System.Windows.Media.Imaging;
using System.IO;
using System.Xml.Linq;
using System.IO.IsolatedStorage;
using Microsoft.Phone.Shell;

namespace PushNotifications
{
  public partial class MainPage : PhoneApplicationPage
  {
    private HttpNotificationChannel httpChannel;
    const string channelName = "WeatherUpdatesChannel";
    const string fileName = "PushNotificationsSettings.dat";
    const int pushConnectTimeout = 120;

    public MainPage()
    {
      InitializeComponent();

      SupportedOrientations = SupportedPageOrientation.Portrait | SupportedPageOrientation.Landscape;

      if (!TryFindChannel())
        DoConnect();
    }

    #region Channel event handlers
    void httpChannel_ExceptionOccurred(object sender, NotificationChannelExceptionEventArgs e)
    {
      UpdateStatus("Error occurred: " + e.Exception.Message);
    }

    void httpChannel_HttpNotificationReceived(object sender, HttpNotificationEventArgs e)
    {
      Trace("===============================================");
      Trace("RAW notification arrived:");

      string weather, location, temperature;
      ParseRAWPayload(e.Notification.Body, out weather, out location, out temperature);

      Dispatcher.BeginInvoke(() => this.textBlockListTitle.Text = location);
      Dispatcher.BeginInvoke(() => this.txtTemperature.Text = temperature);
      Dispatcher.BeginInvoke(() => this.imgWeatherConditions.Source = new BitmapImage(new Uri(@"Images/" + weather + ".png", UriKind.Relative)));
      Trace(string.Format("Got weather: {0} with {1}F at location {2}", weather, temperature, location));

      Trace("===============================================");
    }

    void httpChannel_ChannelUriUpdated(object sender, NotificationChannelUriEventArgs e)
    {
      Trace("Channel opened. Got Uri:\n" + httpChannel.ChannelUri.ToString());
      Dispatcher.BeginInvoke(() => SaveChannelInfo());
      Trace("Subscribing to channel events");
      SubscribeToService();
      SubscribeToNotifications();
      Dispatcher.BeginInvoke(() => UpdateStatus("Channel created successfully"));
    }

    void httpChannel_ShellNotificationReceived(object sender, NotificationEventArgs e)
    {
      Trace("===============================================");
      Trace("Toast notification arrived:");
      foreach (var key in e.Collection.Keys)
      {
        string msg = e.Collection[key];

        Trace(msg);
        Dispatcher.BeginInvoke(() => UpdateStatus("Toast message: " + msg));
      }
      Trace("===============================================");
    }

    void httpChannel_ShellEntryPointNotificationReceived(object sender, NotificationEventArgs e)
    {
      Trace("===============================================");
      Trace("Tile notification arrived:");
      foreach (var key in e.Collection.Keys)
      {
        string msg = e.Collection[key];

        Trace(msg);
        Dispatcher.BeginInvoke(() => UpdateStatus("Tile message: " + msg));
      }
      Trace("===============================================");
    }
    #endregion

    #region Misc logic
    private void DoConnect()
    {
      //TODO - place connection logic here
      try
      {
        Trace("Creating new channel");
        //Create the channel
        httpChannel = new HttpNotificationChannel(channelName, "HOLWeatherService");

        Trace("Subscribing to channel events");
        //Register to UriUpdated event - occurs when channel successfully opens
        httpChannel.ChannelUriUpdated += new EventHandler<NotificationChannelUriEventArgs>(httpChannel_ChannelUriUpdated);

        SubscribeToChannelEvents();

        Trace("Opening channel");
        httpChannel.Open();
        UpdateStatus("Channel open requested");

      }
      catch (NotificationChannelExistsException)
      {
        //Cattch existing channel exception
        Trace("Channel exists! Trying to retrieve by ChannelName");
        //Try to find an existing channel
        httpChannel = HttpNotificationChannel.Find(channelName);

        if (null != httpChannel.ChannelUri)
        {
          Trace("Got existing channel");
          SubscribeToChannelEvents();

          Trace("Subscribing to channel events");
          SubscribeToService();
          SubscribeToNotifications();

          UpdateStatus("Channel recovered");
        }
        else
        {
          Trace("Channel not found...");
          httpChannel.Close();
          UpdateStatus("The channel was closed prematurely");
        }
      }
      catch (NotificationChannelOpenException)
      {
        //Cattch open channel exception - emulator bug workaround
        Trace("EMU workaround - wait for " + pushConnectTimeout.ToString() + " seconds to allow emulator connect to the Cloud Push Service");
        //Wait for emulator to connect to the push service and try-again - EMULATOR ONLY!!!
        int elapsedSeconds = pushConnectTimeout;
        DispatcherTimer timer = new DispatcherTimer();
        timer.Interval = new TimeSpan(0, 0, 0, 1, 0);
        timer.Tick += (sender, args) =>
        {
          Dispatcher.BeginInvoke(() => UpdateStatus("Waiting for push service for " + elapsedSeconds.ToString() + " seconds..."));
          if (--elapsedSeconds == 0)
          {
            timer.Stop();
            Dispatcher.BeginInvoke(() => UpdateStatus("Plase restart the application..."));
            Trace("EMU workaround - finish waiting. Restart the application!");
          }
        };
        timer.Start();
      }
      catch (Exception ex)
      {
        UpdateStatus("Channel error: " + ex.Message);
      }
    }

    private void ParseRAWPayload(Stream e, out string weather, out string location, out string temperature)
    {
      XDocument document;

      using (var reader = new StreamReader(e))
      {
        string payload = reader.ReadToEnd().Replace('\0', ' ');
        document = XDocument.Parse(payload);
      }

      location = (from c in document.Descendants("WeatherUpdate")
                  select c.Element("Location").Value).FirstOrDefault();
      Trace("Got location: " + location);

      temperature = (from c in document.Descendants("WeatherUpdate")
                     select c.Element("Temperature").Value).FirstOrDefault();
      Trace("Got temperature: " + temperature);

      weather = (from c in document.Descendants("WeatherUpdate")
                 select c.Element("WeatherType").Value).FirstOrDefault();
    }
    #endregion

    #region Tracing and Status Updates
    private void UpdateStatus(string message)
    {
      txtStatus.Text = message;
    }

    private void Trace(string message)
    {
#if DEBUG
      Debug.WriteLine(message);
#endif
    }
    #endregion

    #region Loading/Saving Channel Info
    private bool TryFindChannel()
    {
      bool bRes = false;

      Trace("Getting IsolatedStorage for current Application");
      using (IsolatedStorageFile isf = IsolatedStorageFile.GetUserStoreForApplication())
      {
        Trace("Checking channel data");
        if (isf.FileExists(fileName))
        {
          Trace("Channel data exists! Loading...");
          using (IsolatedStorageFileStream isfs = new IsolatedStorageFileStream(fileName, FileMode.Open, isf))
          {
            using (StreamReader sr = new StreamReader(isfs))
            {
              string uri = sr.ReadLine();
              Trace("Finding channel");
              httpChannel = HttpNotificationChannel.Find(channelName);

              if (null != httpChannel)
              {
                if (httpChannel.ChannelUri.ToString() == uri)
                {
                  Dispatcher.BeginInvoke(() => UpdateStatus("Channel retrieved"));
                  SubscribeToChannelEvents();
                  SubscribeToNotifications();
                  SubscribeToService();
                  bRes = true;
                }
                sr.Close();
              }
            }
          }
        }
        else
          Trace("Channel data not found");
      }

      return bRes;
    }

    private void SaveChannelInfo()
    {
      Trace("Getting IsolatedStorage for current Application");
      using (IsolatedStorageFile isf = IsolatedStorageFile.GetUserStoreForApplication())
      {
        Trace("Creating data file");
        using (IsolatedStorageFileStream isfs = new IsolatedStorageFileStream(fileName, FileMode.Create, isf))
        {
          using (StreamWriter sw = new StreamWriter(isfs))
          {
            Trace("Saving channel data...");
            sw.WriteLine(httpChannel.ChannelUri.ToString());
            sw.Close();
            Trace("Saving done");
          }
        }
      }
    }
    #endregion

    #region Subscriptions
    private void SubscribeToService()
    {
      //Hardcode for solution - need to be updated in case the REST WCF service address change
      string baseUri = "http://localhost:8000/RegirstatorService/Register?uri={0}";
      string theUri = String.Format(baseUri, httpChannel.ChannelUri.ToString());
      WebClient client = new WebClient();
      client.DownloadStringCompleted += (s, e) =>
      {
        if (null == e.Error)
          Dispatcher.BeginInvoke(() => UpdateStatus("Registration succeeded"));
        else
          Dispatcher.BeginInvoke(() => UpdateStatus("Registration failed: " + e.Error.Message));
      };
      client.DownloadStringAsync(new Uri(theUri));
    }

    private void SubscribeToChannelEvents()
    {
      //Subscribe to the channel events
      httpChannel.HttpNotificationReceived += new EventHandler<HttpNotificationEventArgs>(httpChannel_HttpNotificationReceived);
      httpChannel.ExceptionOccurred += new EventHandler<NotificationChannelExceptionEventArgs>(httpChannel_ExceptionOccurred);
      httpChannel.ShellEntryPointNotificationReceived += new EventHandler<NotificationEventArgs>(httpChannel_ShellEntryPointNotificationReceived);
      httpChannel.ShellNotificationReceived += new EventHandler<NotificationEventArgs>(httpChannel_ShellNotificationReceived);
    }

    private void SubscribeToNotifications()
    {
      try
      {
        Trace("Registering to Toast Notifications");
        httpChannel.BindToShellNotification();
      }
      catch (NotificationChannelExistsException)
      {
        //Nothing to do - channel already exists and subscribed
      }

      try
      {
        Trace("Registering to Tile Notifications");
        ShellEntryPoint shellEntryPoint = new ShellEntryPoint();
        shellEntryPoint.RemoteImageUri = new Uri("http://developer.windowsphone.com/Common/Fire/Images/bg-brand.png", UriKind.Absolute);

        httpChannel.BindToShellEntryPoint(shellEntryPoint);
      }
      catch (NotificationChannelExistsException)
      {
        //Nothing to do - channel already exists and subscribed
      }
    }
    #endregion
  }
}